import sys
from pathlib import Path
from dataclasses import dataclass
from typing import Any

import typer
from rich import print

from .commands import provision, build, deploy, info, shell, toolbox, update
from .commands.template_editor import inspector
from .commands.telemetry import send_event
from .models import Application
from .check_dependencies import check_deps
from cnc.logger import get_logger

log = get_logger(__name__)

app = typer.Typer()
app.add_typer(provision.app, name="provision")
app.add_typer(build.app, name="build")
app.add_typer(deploy.app, name="deploy")
app.add_typer(inspector.app, name="inspector")
app.add_typer(info.app, name="info")
app.add_typer(shell.app, name="shell")
app.add_typer(toolbox.app, name="toolbox")
app.add_typer(update.app, name="update")

@dataclass
class Common:
    application: Any = None
    collection: Any = None


@app.callback()
def add_common(
    ctx: typer.Context,
    config_file_path: Path = typer.Option(
        default="cnc.yml", envvar="CNC_CONFIG_PATH", help="CNC config file path"
    ),
    environments_file_path: Path = typer.Option(
        default="environments.yml",
        envvar="CNC_ENVIRONMENTS_PATH",
        help="Environments data file path",
    ),
):
    """Common Entry Point for whole app"""
    if "--help" in sys.argv:
        return True

    check_deps()

    if not config_file_path.is_file():
        log.error(f"Config file not found: {config_file_path}")
        raise typer.Exit(code=1)

    if not environments_file_path.is_file():
        log.error(f"Environments data file not found: {environments_file_path}")
        raise typer.Exit(code=1)

    application = Application.from_environments_yml(
        environments_file_path,
        config_file_path,
    )
    if not application:
        log.error(f"No application for: {environments_file_path}")
        raise typer.Exit(code=1)

    ctx.obj = Common(application)


@app.command()
def hello():
    send_event("hello")
    print("[bold red]Welcome[/bold red] to [bold purple]" "CNC[/bold purple]")

    raise typer.Exit()


if __name__ == "__main__":
    app()



import yaml
from pathlib import Path
from typing import Union, List, ClassVar, Optional
from pydantic import (
    Field,
    model_validator,
    field_validator,
)

from .base_model import BaseModel
from .providers.amazon.environment_collection import AWSEnvironmentCollection
from .providers.google.environment_collection import GCPEnvironmentCollection
from cnc.utils import clean_name_string

from cnc.logger import get_logger

log = get_logger(__name__)


class TemplateConfig(BaseModel):
    template_directory: Optional[str] = None
    provision_filename: Optional[str] = "main.tf.j2"
    deploy_filename: Optional[str] = "main.sh.j2"
    build_filename: Optional[str] = "main.sh.j2"

    @property
    def enabled(self):
        return bool(self.template_directory)


class Application(BaseModel):
    name: str
    provider: str
    region: str
    flavor: str
    version: Union[int, float]
    collections: List[Union[AWSEnvironmentCollection, GCPEnvironmentCollection]] = (
        Field(discriminator="provider")
    )
    template_config: Optional[TemplateConfig] = Field(default_factory=TemplateConfig)

    GCP_APP_PROVIDER: ClassVar[str] = "gcp"
    AWS_APP_PROVIDER: ClassVar[str] = "aws"

    # ------------------------------
    # Validators
    # ------------------------------
    @model_validator(mode="before")
    def ensure_collection_provider(cls, data):
        for collection in data.get("collections", []):
            collection["provider"] = data.get("provider")

        return data

    @model_validator(mode="after")
    def annotate_children(self):
        for collection in self.collections:
            collection.application = self

            for environment in collection.environments:
                environment.application = self

        return self

    @field_validator("name", mode="before")
    def validate_name(cls, value: str) -> str:
        return clean_name_string(value)

    # ------------------------------
    # Class methods
    # ------------------------------
    @classmethod
    def from_environments_yml(
        cls, data_file_path: Path, config_file_path: Path = "cnc.yml"
    ):
        with open(data_file_path) as parsed_data:
            env_yml_data = yaml.safe_load(parsed_data)
        return cls.model_validate(
            env_yml_data,
            context={
                "config_file_path": str(config_file_path),
                "provider": env_yml_data.get("provider"),
            },
        )

    @classmethod
    def from_environments_data(cls, data):
        return cls.model_validate(
            data,
            context={"provider": data.get("provider")},
        )

    # ------------------------------
    # Properties
    # ------------------------------
    @property
    def default_collection(self):
        for collection in self.collections:
            if collection.default:
                return collection

        log.debug(
            f"no default provided - returning first collection as default for {self}"
        )
        # only happens if no default defined
        return self.collections[0]

    @property
    def provider_is_gcp(self):
        return self.provider == self.GCP_APP_PROVIDER

    @property
    def provider_is_aws(self):
        return self.provider == self.AWS_APP_PROVIDER

    @property
    def config_files_path(self):
        return (
            f"{self.config_files_parent_dir}/app_{self.name}/"
            f"{self.provider}_{self.flavor}/{self.version}"
        )

    @property
    def config_files_parent_dir(self):
        return f"/tmp/.cnc_tmp_{self.name}"

    @property
    def environments(self):
        if not hasattr(self, "_envs"):
            self._envs = []

        if not self._envs:
            for collection in self.collections:
                for e in collection.environments:
                    self._envs.append(e)

        return self._envs

    # ------------------------------
    # Instance methods
    # ------------------------------
    def __repr__(self):
        return (
            f"<Application (name: {self.name} | provider: {self.provider} "
            f"({self.flavor}/{self.version}))>"
        )

    __str__ = __repr__

    def collection_by_name(self, collection_name=None):
        if collection_name:
            for collection in self.collections:
                if collection.name == collection_name:
                    return collection
        else:
            return self.default_collection



import subprocess
import os
import hashlib
import json
import time
import base64
import re

import pygohcl
from .cycle_stage_base import CollectionTemplatedBase

from cnc.logger import get_logger

log = get_logger(__name__)


class ProvisionStageManager(CollectionTemplatedBase):
    template_type = "provision"
    entrypoint_script_name = "main.tf.j2"

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.output_only = kwargs.get("output_only", False)

    @property
    def tf_state_namespace(self):
        return f"{self.collection.cloud_resource_namespace}"

    @property
    def infra_state_hash(self):
        # this is the template, not the rendered TF file
        main_tpl = f"{self.config_files_path}/main.tf.j2"
        if not os.path.isfile(main_tpl):
            log.info(f"{main_tpl} not a file...")
            return ""

        _hash = hashlib.sha256()
        with open(main_tpl) as f:
            _hash.update(f.read().encode())
        template_hash = _hash.hexdigest()

        _hash = hashlib.sha256()
        _state = f"{self.collection.unique_id}:{template_hash}:{self.application.version}:{self.application.flavor}:{self.collection.account_id}:{self.application.name}"
        # log.debug(f"Raw State for {self}: {_state}")
        hashed = hashlib.sha256()
        hashed.update(_state.encode())
        hexdigest = hashed.hexdigest()
        # log.info(f"Digest: {hexdigest}")
        return hexdigest

    def make_ready_for_use(
        self,
        should_cleanup=True,
        should_regenerate_config=True,
    ):
        if should_cleanup:
            log.debug(f"Cleaning up & setting up at start for {self}")

            if not self.cleanup():
                return False

            if not self.setup():
                return False

        if should_regenerate_config:
            log.debug(
                f"Writing {self.application.template_config.provision_filename} for {self}"
            )
            if not self.write_template(
                self.application.template_config.provision_filename
            ):
                return False

        return self.init()

    def template_context(self):
        if self.application.provider_is_aws:
            frontend_names = []
            for svc in self.collection.frontend_services:
                if svc.name not in frontend_names:
                    frontend_names.append(svc.name)

            # this hash is used to prevent tf errors
            # in the event of a major infra change
            # (e.g. switching from a frontend only app to a backend only app)
            #
            # when the hash changes, it forces creation of a new
            # cloudfront dist instead of updating an existing one in place
            fe_hash = base64.b64encode("".join(frontend_names).encode()).decode()
            fe_hash = re.sub("\\W", "", fe_hash)
        else:
            fe_hash = ""

        return {
            "output_only": self.output_only,
            "app": self.application,
            "env_collection": self.collection,
            "config_renderer": self,
            "frontend_hash": fe_hash,
            "svc_path_lambda": lambda svc: len(svc.settings.url_path),
            "sorted": sorted,
            "has_postgres_db": [
                True
                for r in self.collection.database_resources
                if r.settings.engine == "postgres"
            ],
            "has_mysql_db": [
                True
                for r in self.collection.database_resources
                if r.settings.engine == "mysql"
            ],
            "has_mssql_db": [
                True
                for r in self.collection.database_resources
                if "sqlserver" in r.settings.engine
            ],
        }

    def copy_templates(self):
        self.copy_template_dir()

        if not self.output_only:
            self.collection.generate_tf_assets(
                self.config_files_path, self.rendered_files_path
            )

    @property
    def config_files_path(self):
        if self.output_only:
            return (
                f"{self.collection.config_files_path}/output_only/{self.template_type}"
            )

        return f"{self.collection.config_files_path}/{self.template_type}"

    @property
    def is_ready(self):
        return os.path.isdir(f"{self.rendered_files_path}/.terraform")

    def init(self):
        log.debug(f"Installing TF modules/providers for {self}")
        _ret = self._tf_command("init", capture_output=True)
        return self._return_true_if_successful(_ret)

    def unlock(self, lock_id):
        _ret = self._tf_command("force-unlock", [str(lock_id)], options_list=["-force"])
        return self._return_true_if_successful(_ret)

    def saved_plan_exists(self, plan_filename="tfplan"):
        filename = f"{self.rendered_files_path}/{plan_filename}"
        return os.path.isfile(filename)

    def clean_saved_plan(self, plan_filename="tfplan"):
        filename = f"{self.rendered_files_path}/{plan_filename}"
        if self.saved_plan_exists(plan_filename=plan_filename):
            os.remove(filename)

    def plan(self, save=False, plan_filename="tfplan", target=None):
        args = ["-json"]

        if save:
            self.clean_saved_plan(plan_filename)
            args.extend(["-out", plan_filename])
        if target:
            args.extend(["-target", target])

        _ret = self._tf_command("plan", args=args, capture_output=True)

        if not _ret:
            log.info(f"{self} got None in plan")
            return {"status": "not_ready"}

        for _log in _ret.stdout.decode().splitlines():
            try:
                parsed_log = json.loads(_log)
            except Exception:
                log.debug(f"Cannot parse log line from TF ({self}): {_log}")
                continue
            if parsed_log.get("type") == "change_summary":
                parsed_log.update({"status": "ready"})
                return parsed_log
            elif (
                parsed_log.get("@message", "")
                == "Error: Error acquiring the state lock"
            ):
                # this is just so we log the lock ID - would be best to parse it and return in the status dict...
                _ret = self._tf_command("plan", capture_output=True)
                lock_logs = _ret.stderr.decode()
                log.debug(f"{lock_logs}")
                return {"status": "locked", "info": lock_logs}

        log.debug(f"No change summary log available for {self}: \n...")

    def plan_json(self, plan_filename="tfplan"):
        # log.debug(f"Getting plan JSON for {self}: {plan_filename}")
        _ret = self._tf_command("show", ["-json", plan_filename], capture_output=True)

        try:
            return json.loads(_ret.stdout)
        except Exception:
            log.debug(f"Cannot parse plan_json from TF ({self}): {_ret}")
            return {}

    def destroy_item(self, item):
        return self._tf_command("destroy", args=["-auto-approve", "-target", item])

    def destroy(self):
        return self._tf_command("destroy", args=["-auto-approve"], capture_output=True)

    def validate(self):
        _ret = self._tf_command("validate")
        return self._return_true_if_successful(_ret)

    def parse_config_file(self, tfconfig="main.tf"):
        with open(f"{self.rendered_files_path}/{tfconfig}", "r") as f:
            parsed = pygohcl.loads(f.read())
        return parsed

    @property
    def summarized_resource_state(self):
        state = {}
        parsed = self.parse_config_file()

        for resource_type, resources in parsed["resource"].items():
            for address, resource_data in resources.items():
                if resource_type in state:
                    state[resource_type].append({"address": address})
                else:
                    state[resource_type] = [{"address": address}]

        return state

    def apply(self, use_saved=False, plan_filename="tfplan"):
        args = ["-json"]
        if use_saved:
            args.extend([plan_filename])
        _ret = self._tf_command(
            "apply", args=args, options_list=["-auto-approve"], capture_output=True
        )

        if not _ret:
            log.warning(f"_ret is None for {self}")
            return {}

        cloudfront_error = ""
        if _ret.returncode != 0:
            output = _ret.stdout.decode()
            for _line in output.splitlines():
                try:
                    line = json.loads(_line or "{}")
                except Exception:
                    log.info(f"Cannot load TF log line: {_line}")
                    continue

                imported = False
                for stmt in ["Error 1007: Can't create database", "already exists"]:
                    if not imported and stmt in line.get("@message", ""):
                        imported = self.import_existing_resource(line)

                if imported:
                    continue

                line_msg = line.get("@message", "")
                if "Error 400: Billing account" in line_msg:
                    return {"error": "enable_billing"}
                elif re.match(".+cloudfront_distribution.+Creation errored", line_msg):
                    cloudfront_error = f"Error: {line_msg}"
                else:
                    log_noconflict_tf_line(line)

        try:
            parsed_log = json.loads(_ret.stdout.decode().splitlines()[-1])

            if cloudfront_error:
                if parsed_log.get("@message"):
                    parsed_log["@message"] += f" {cloudfront_error}"
                else:
                    parsed_log["@message"] = cloudfront_error

            return parsed_log
        except Exception:
            log.debug(
                f"Cannot parse apply response logs from TF ({self}): {_ret.stdout}"
            )
            return {"error": f"Cannot parse apply response logs: {_ret.stdout}"}

    def import_existing_resource(self, tf_log_line):
        address = tf_log_line.get("diagnostic", {}).get("address")
        if not address:
            log.info(f"No address to import for existing resource: {tf_log_line}")
            return False

        resource_id = None
        if "Error creating Database" in tf_log_line.get("@message", ""):
            if self.collection.is_production:
                log.info(f"Skipping database import for production for {tf_log_line}")
                return False

            db_name = self.parse_database_resource_id(tf_log_line)
            db_instance_name = ""
            for db_resource in self.collection.database_resources:
                if db_resource.instance_name in address:
                    db_instance_name = db_resource.instance_name
                    break

            resource_id = f"{db_instance_name}/{db_name}"

        if not resource_id:
            log.info(f"No resource ID to import for {tf_log_line}")
            return False

        log.info(f"Trying importing {resource_id} to {address}")
        return self.import_resource(address, resource_id)

    def parse_database_resource_id(self, tf_log_line):
        msg = tf_log_line.get("@message", "")

        try:
            return re.search('".+"', msg)[0].replace('"', "")
        except Exception:
            log.info(f"Cannot parse DB name from {msg}")

    def debug(self):
        with open(
            f"{self.rendered_files_path}/{self.application.template_config.provision_filename}",
            "r",
        ) as f:
            print(f.read())

    def import_resource(self, resource_address, resource_id):
        _ret = self._tf_command("import", args=[resource_address, resource_id])
        return self._return_true_if_successful(_ret)

    def list_state(self):
        _ret = self._tf_command("state", args=["list"])
        return self._return_true_if_successful(_ret)

    def clean_state(self, resource_address):
        # TODO: make sure it's a valid address in the list?
        _ret = self._tf_command("state", args=["rm", resource_address])
        return self._return_true_if_successful(_ret)

    def output(self):
        _ret = self._tf_command("output", ["-json"], capture_output=True)
        if _ret and (_ret.returncode == 0):
            return json.loads(_ret.stdout.decode())
        return {}

    def _return_true_if_successful(self, _ret):
        if not _ret:
            return False

        if _ret.returncode == 0:
            return True
        else:
            # self.debug()
            return False

    def _tf_command(self, command, args=None, options_list=None, capture_output=False):
        start = int(time.time())
        if (not self.is_ready) and (command != "init"):
            log.info(f"Cannot {command}|{options_list} - not ready!: {self}")
            if command not in ["init"]:
                return

        full_command = ["terraform", command]

        if options_list and isinstance(options_list, list):
            full_command = full_command + options_list
        if args:
            full_command = full_command + args

        tf_cache_dir = "/tmp/CNC_tfplugin"
        if not os.path.isdir(tf_cache_dir):
            os.mkdir(tf_cache_dir)

        _ret = subprocess.run(
            full_command,
            cwd=self.rendered_files_path,
            capture_output=capture_output,
            env=dict(os.environ, TF_PLUGIN_CACHE_DIR=tf_cache_dir),
        )

        end = int(time.time())
        duration = end - start

        log.info(
            f"TF RUN ({self}): [{full_command}] {_ret.returncode} in {duration} seconds"
        )

        if (_ret.returncode != 0) and (command == "init"):
            log.debug(f"dir contents: {os.listdir(self.rendered_files_path)}")
            log.debug(_ret.stdout)
            log.debug(_ret.stderr)
            log.debug("\n\n----------------\n\n")
            # self.debug()
            log.debug("\n\n----------------\n\n")

        return _ret


def log_noconflict_tf_line(line):
    ignore_strings = [
        "Still creating...",
        "Still destroying...",
        "Warning: Applied changes may be incomplete",
        "Destroying...",
        "Plan to delete",
    ]

    for _ in ignore_strings:
        message = line.get("@message", "")
        if (message == "Terraform 1.1.4") or (_ in message):
            return

    log.info(f"TF log: {line}")



import re
import hashlib
from typing import List, Optional, Union, Any

from pydantic import BaseModel, field_validator, Field, model_validator

from cnc.constants import EnvironmentVariableTypes

# from cnc.constants import EnvironmentVariableDestinations
from ..environment_variable import EnvironmentVariable
from .resource import (
    CacheResourceSettings,
    QueueResourceSettings,
    BucketResourceSettings,
)
from .service import Service, ProviderDatabaseResourceSettings

from cnc.logger import get_logger

log = get_logger(__name__)


class PlatformBuildSettings(BaseModel):
    machine_type: str


class BuildSettings(BaseModel):
    platform_settings: Optional[PlatformBuildSettings] = {}


class ResourceService(Service):
    settings: Union[
        ProviderDatabaseResourceSettings,
        CacheResourceSettings,
        BucketResourceSettings,
        QueueResourceSettings,
    ] = Field(alias="x-cnc", discriminator="type")


class InternalSettings(BaseModel):
    build_settings: Optional[BuildSettings] = {}
    preview_inactivity_timeout_hours: Optional[int] = 90 * 24
    explicit_resources: Optional[List[ResourceService]] = Field(
        default_factory=list,
        alias="resources",
    )

    @field_validator("explicit_resources", mode="before")
    def format_resources(cls, value: list):
        formatted_resource_svcs = []
        for resource in value:
            resource_name = resource["name"]
            del resource["name"]

            formatted_resource_svcs.append(
                {
                    "name": resource_name,
                    "x-cnc": resource,
                }
            )

        return formatted_resource_svcs


class AppConfig(BaseModel):
    settings: Optional[InternalSettings] = Field(
        default_factory=InternalSettings, alias="x-cnc"
    )
    services: List[Service]

    # ------------------------------
    # Parent relationships
    # ------------------------------
    environment: Optional[Any] = None

    def __repr__(self):
        return f"<AppConfig: {[s.name for s in self.services]}>"

    __str__ = __repr__

    @field_validator("services", mode="before")
    def validate_services(cls, value: dict) -> list[dict]:
        with_names = []
        for k, v in value.items():
            if not isinstance(v, dict):
                raise ValueError(
                    f"({k} is bad compose service - service values must be dictionaries!"
                )

            # this is to ensure we ignore config from the compose file
            v.update({"name": k})
            if v.get("config"):
                del v["config"]

            with_names.append(v)
        return with_names

    @model_validator(mode="after")
    def annotate_children(self):
        for service in self.services + self.settings.explicit_resources:
            service.config = self
            service.settings.config = self

        return self

    @property
    def resources(self):
        all_resources = []

        for service in self.services:
            if service.settings.type in ["database", "cache"]:
                all_resources.append(service)

        all_resources.extend(self.settings.explicit_resources)

        return all_resources

    @property
    def unique_id(self):
        hash_strings = []

        for service in self.services:
            hash_strings.append(service.settings.unique_id)

        _hash = hashlib.sha256()
        _hash.update(":".join(sorted(hash_strings)).encode())
        return _hash.hexdigest()

    @property
    def has_backend(self):
        return any(s.is_backend for s in self.services)

    @property
    def build_machine_type(self):
        build_settings = self.settings.build_settings
        if build_settings:
            if build_settings.platform_settings:
                return build_settings.platform_settings.machine_type

    def filtered_environment_items(
        self, service_name=None, variable_type=None, pattern=None
    ):
        if (
            variable_type
            and variable_type not in EnvironmentVariableTypes.allowed_types()
        ):
            raise ValueError(
                f"Invalid variable type in filtered_environment_items: {variable_type}"
            )

        regex_vars_by_name = []
        for item in self.environment.environment_variables:
            if pattern:
                if re.match(pattern, item.name):
                    regex_vars_by_name.append(item)
            else:
                regex_vars_by_name.append(item)

        service_vars = []
        for env_item in regex_vars_by_name:
            if service_name is None:
                service_vars.append(env_item)
            elif (env_item.service is None) or (env_item.service == service_name):
                service_vars.append(env_item)

        type_vars = []
        if variable_type:
            for item in service_vars:
                if variable_type == item.variable_type:
                    type_vars.append(item)
                elif (
                    variable_type == EnvironmentVariableTypes.VARIABLE_TYPE_SECRET
                    and item.alias
                    and item.secret_id
                ):
                    type_vars.append(item)
                elif (
                    variable_type == EnvironmentVariableTypes.VARIABLE_TYPE_OUTPUT
                    and item.alias
                    and item.output_name
                ):
                    type_vars.append(item)
                elif (
                    variable_type == EnvironmentVariableTypes.VARIABLE_TYPE_STANDARD
                    and item.alias
                    and not item.output_name
                    and not item.secret_id
                ):
                    type_vars.append(item)
        else:
            type_vars = service_vars

        return type_vars

    @property
    def reserved_environment_item_names(self):
        all_names = []

        for item in self.managed_environment_variables:
            all_names.append(item.name)
        for item in self.managed_environment_secrets:
            all_names.append(item.name)

        return set(all_names)

    @property
    def managed_environment_variables(self):
        _vars = {
            "CNC_APPLICATION_NAME": self.environment.application.name,
            "CNC_ENVIRONMENT_NAME": self.environment.name,
            "CNC_ENVIRONMENT_DOMAIN": self.environment.domains[0],
            "CNC_ENVIRONMENT_REGION": self.environment.collection.region,
        }

        for service in self.services:
            _vars.update(service.settings.managed_environment_variables)

        return [self.variable_object({"name": k, "value": v}) for k, v in _vars.items()]

    @property
    def managed_environment_secrets(self):
        _vars = {}

        for resource in self.resources:
            _vars.update(resource.settings.managed_environment_secrets)

        return [
            self.variable_object({"name": k, "secret_id": v}) for k, v in _vars.items()
        ]

    @property
    def database_resources(self):
        instances = []

        for resource in self.resources:
            if resource.is_database:
                instances.append(resource)

        return instances

    @property
    def cache_resources(self):
        instances = []

        for resource in self.resources:
            if resource.is_cache:
                instances.append(resource)

        return instances

    @property
    def object_storage_resources(self):
        instances = []

        for resource in self.resources:
            if resource.is_object_storage:
                instances.append(resource)

        return instances

    def variable_object(self, data):
        variable = EnvironmentVariable.model_validate(data)
        variable.environment = self.environment
        variable.collection = self.environment.collection
        return variable



import os
import re
import subprocess
import shutil
from pathlib import Path
from functools import partial, cached_property

from jinja2 import (
    Environment,
    FileSystemLoader,
    select_autoescape,
    StrictUndefined,
)

from cnc.logger import get_logger

log = get_logger(__name__)


class _TemplatedBase:
    template_type = "build"
    entrypoint_script_name = "main.sh.j2"

    def __repr__(self):
        return f"<{self.__class__.__name__} @ {self.working_dir} -> {self.config_files_path}>"

    __str__ = __repr__

    @property
    def config_files_path(self):
        return f"{self.collection.config_files_path}/{self.template_type}"

    @property
    def rendered_files_path(self):
        return f"{self.config_files_path}/_cnc_output"

    @property
    def custom_template_dir(self):
        return f"{self.config_files_path}/custom"

    @property
    def build_args(self):
        return {}

    def setup(self):
        if not os.path.isdir(self.config_files_path):
            os.makedirs(self.config_files_path, exist_ok=True)

        if not os.path.isdir(self.rendered_files_path):
            os.makedirs(self.rendered_files_path, exist_ok=True)

        if self.template_config.enabled:
            if not os.path.isdir(self.custom_template_dir):
                os.makedirs(self.custom_template_dir, exist_ok=True)

        self.copy_templates()
        return True

    def copy_template_dir(self):
        src_dir = Path(__file__).parent.parent
        cwd = self.working_dir

        # Path to the directory where templates are finally consolidated
        final_template_dir = self.config_files_path

        # Copy shared default templates
        shared_template_dir = (
            f"{src_dir}/flavors/{self.application.provider}"
            f"/shared/{self.template_type}"
        )
        if os.path.isdir(shared_template_dir):
            shutil.copytree(
                shared_template_dir,
                final_template_dir,
                dirs_exist_ok=True,
            )

        # Copy flavor-specific templates, which can overwrite shared defaults
        flavor_dir = (
            f"{src_dir}/flavors/{self.application.provider}/"
            f"{self.application.flavor}/{self.application.version}"
            f"/{self.template_type}"
        )
        if os.path.isdir(flavor_dir):
            shutil.copytree(
                flavor_dir,
                final_template_dir,
                dirs_exist_ok=True,
            )

        # Copy custom templates if enabled, which can overwrite any of the defaults
        if self.template_config.enabled:
            custom_template_dir = Path(
                f"{cwd}/{self.template_config.template_directory}"
            )
            main_file = getattr(
                self.template_config, f"{self.template_type}_filename", None
            )

            if not custom_template_dir.is_dir():
                raise Exception(
                    f"Template directory {custom_template_dir} does not exist"
                )

            shutil.copytree(
                str(custom_template_dir),
                self.custom_template_dir,
                dirs_exist_ok=True,
            )

            # Check if the custom main template file exists
            path_to_copy = Path(
                f"{custom_template_dir}/{self.template_type}/{main_file}"
            )
            if path_to_copy.is_file():
                # Optionally delete an entrypoint script in the final directory before copying the main file
                entry_script_path = Path(
                    f"{final_template_dir}/{self.entrypoint_script_name}"
                )
                if entry_script_path.exists():
                    os.remove(entry_script_path)
                shutil.copy(
                    str(path_to_copy),
                    f"{str(final_template_dir)}/{self.entrypoint_script_name}",
                )

        return True

    def get_template(self, name, template_directory=None):
        template_directory = template_directory or self.config_files_path
        env = Environment(
            loader=FileSystemLoader(template_directory),
            autoescape=select_autoescape(),
            undefined=StrictUndefined,
        )

        try:
            return env.get_template(name)
        except Exception as e:
            log.warning(f"Could not get template ({name}): {e}")
            self.debug_template_output_directory(template_directory)
            raise e

    def cleanup(self):
        shutil.rmtree(
            self.config_files_path,
            ignore_errors=True,
        )
        return True

    def template_context(self):
        return {}

    def write_template(
        self,
        name,
        output_name=None,
        context=None,
        additional_context=None,
        template_directory=None,
    ):
        template = self.get_template(name, template_directory)
        if not template:
            log.info(f"No template for {name} for {self}")
            return

        output_name = output_name or ".".join(name.split(".")[:-1])
        output_path = f"{self.rendered_files_path}/{output_name}"
        with open(output_path, "w") as f:
            # render template
            template_context = context or self.template_context()
            template_context.update(additional_context or {})
            rendered = template.render(**template_context)
            # strip excess newlines
            cleaned = re.sub(r"\n\s*\n", "\n", rendered)
            f.write(cleaned)

        return True

    def _wrapped_write_template(self, *args, **kwargs):
        self.write_template(*args, **kwargs)
        return ""

    def write_template_with_context(self, service):
        return partial(
            self._wrapped_write_template, context=self.template_context(service)
        )

    def debug_template_output(self, output_name):
        try:
            print(
                f"\n\n---------CONTENTS of {output_name}---------\n\n"
                + open(f"{self.config_files_path}/_cnc_output/{output_name}").read()
                + f"\n\n---------END OF CONTENTS for {output_name}---------\n\n"
            )
        except FileNotFoundError as e:
            log.error(f"File {output_name} not found: {e}")
            return False

    def debug_template_output_directory(self, template_directory=None):
        template_directory = (
            template_directory or f"{self.config_files_path}/_cnc_output"
        )
        log.debug(f"\noutput directory is: {template_directory}\n")
        subprocess.run(["ls", "-alhR", f"{template_directory}"])

    def file_exists(self, file_path="Dockerfile"):
        cwd = os.getcwd()
        return os.path.isfile(f"{cwd}/{file_path}")


class CollectionTemplatedBase(_TemplatedBase):

    def __init__(
        self,
        collection,
    ):
        self.application = collection.application
        self.collection = collection
        self.template_config = self.application.template_config
        self.working_dir = os.getcwd()

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.collection} | output_only: {self.output_only}>"

    __str__ = __repr__

    @cached_property
    def environment_items(self):
        _all = {}

        for environment in self.collection.environments:
            for item in environment.environment_items:
                _all.update({item.name: item.value})

        return _all


class EnvironmentTemplatedBase(_TemplatedBase):

    def __init__(
        self,
        environment,
        service_tags=None,
    ):
        self.environment = environment

        if isinstance(service_tags, list):
            service_tags = {
                tag.split("=")[0]: tag.split("=")[1] for tag in service_tags
            }

        self.service_tags = service_tags or {}
        self.collection = environment.collection
        self.application = environment.application
        self.template_config = self.application.template_config
        self.scripts_to_run = []

        self.working_dir = os.getcwd()

    def __repr__(self):
        return f"<{self.__class__.__name__}:{self.environment} @ {self.service_tags}>"

    __str__ = __repr__

    @cached_property
    def environment_items(self):
        _all = {}

        for item in self.environment.environment_items:
            _all.update({item.name: item.value})

        return _all

    def copy_templates(self):
        self.copy_template_dir()

    def tag_for_service(self, service_name=None):
        return self.service_tags.get(service_name, "latest")



import re
import hashlib
from typing import (
    List,
    Optional,
    Any,
    Literal,
    Union,
    Annotated,
)
from pydantic import (
    field_validator,
    Field,
    model_validator,
)

from cnc.models.base_model import BaseModel, IgnoredType
from cnc.models.custom_header import CustomHeaders
from cnc.models.providers.amazon import (
    AWSCustomHeaders,
    AWSDeployResourceLimits,
)
from cnc.models.providers.google import GCPDeployResourceLimits
from .utils import validate_command_list
from cnc.utils import clean_name_string

from cnc.logger import get_logger

log = get_logger(__name__)


class ServiceBuildTimeoutsConfig(BaseModel):
    default: Optional[int] = 20
    test: Optional[int] = 20
    migrate: Optional[int] = 20
    snapshot: Optional[int] = 20
    seed: Optional[int] = 20
    deploy: Optional[int] = 20


class ServiceBuildConfig(BaseModel):
    included_paths: Optional[List[str]] = []
    excluded_paths: Optional[List[str]] = []
    existing_image_tag_mode: Optional[Literal["static", "sha"]] = "static"


class PlatformSettings(BaseModel):
    min_scale: Optional[int] = 1
    max_scale: Optional[int] = 4


class SystemSettings(BaseModel):
    health_check: Optional[str] = "/"
    platform_settings: Optional[PlatformSettings] = Field(
        default_factory=PlatformSettings
    )


class CDNSettings(BaseModel):
    enabled: Optional[bool] = True


class GCPCustomHeaders(CustomHeaders):
    provider: Literal["gcp"]


class BaseServiceSettings(BaseModel):
    type: str
    data: Optional[dict] = {}
    url_path: Optional[str] = "/"
    custom_headers: Optional[
        Union[
            AWSCustomHeaders,
            GCPCustomHeaders,
        ]
    ] = Field(
        discriminator="provider",
        default=None,
    )
    cdn: Optional[CDNSettings] = Field(default_factory=CDNSettings)
    system: Optional[SystemSettings] = Field(default_factory=SystemSettings)
    build_config: Optional[ServiceBuildConfig] = Field(
        default_factory=ServiceBuildConfig
    )
    timeouts: Optional[ServiceBuildTimeoutsConfig] = Field(
        default_factory=ServiceBuildTimeoutsConfig
    )
    provider: str

    # ------------------------------
    # Parent relationships
    # ------------------------------
    config: Optional[IgnoredType] = IgnoredType()
    service: Optional[IgnoredType] = IgnoredType()

    # ------------------------------
    # Validators
    # ------------------------------
    @model_validator(mode="before")
    def format_custom_headers(cls, data, info):
        if isinstance(data.get("custom_headers"), list):
            data["custom_headers"] = {
                "headers": data.get("custom_headers", []),
                "provider": info.context.get("provider"),
            }
        return data

    @field_validator("url_path", mode="before")
    def validate_command(cls, value: str) -> str:
        return value.strip().rstrip("/")

    # ------------------------------
    # Properties
    # ------------------------------
    @property
    def is_web(self):
        return False

    @property
    def is_cache(self):
        return False

    @property
    def application(self):
        return self.service.environment.application

    @property
    def is_resource(self):
        return False

    @property
    def unique_id(self):
        _hash = hashlib.sha256()
        _hash.update(
            f"{self.service.environment.application.name}:{self.service.environment.collection.name}:{self.service.environment.name}:{self.service.name}:{self.url_path}:{self.system.health_check}".encode()
        )
        return _hash.hexdigest()

    @property
    def managed_environment_variables(self):
        return {"CNC_INSTANCE_NAME": self.service.instance_name}

    @property
    def custom_request_headers(self):
        headers = []

        if self.custom_headers:
            for header in self.custom_headers.headers:
                if header.header_type == "request":
                    headers.append(header)

        return headers

    @property
    def custom_response_headers(self):
        headers = []

        if self.custom_headers:
            for header in self.custom_headers.headers:
                if header.header_type == "response":
                    headers.append(header)

        return headers

    @property
    def existing_instance_name(self):
        return


class CORSSettings(BaseModel):
    allowed_origins: List[str]


class FrontendCDNConfig(BaseModel):
    enabled: Optional[bool] = False


class BackendCDNConfig(BaseModel):
    enabled: Optional[bool] = False


ProviderDeployResourceLimits = Annotated[
    Union[GCPDeployResourceLimits, AWSDeployResourceLimits],
    Field(discriminator="provider"),
]


class FrontendServiceSettings(BaseServiceSettings):
    type: Literal["frontend"]
    assets_path: Optional[str] = "build"
    build: List[str] = Field(default_factory=list)
    cors: Optional[CORSSettings] = None
    index_file_name: Optional[str] = "index.html"
    cdn: Optional[FrontendCDNConfig] = Field(default_factory=FrontendCDNConfig)

    @field_validator("build", mode="before")
    def validate_command(cls, value: Any) -> list[str]:
        return validate_command_list(value)

    @property
    def is_web(self):
        return True


class Worker(BaseModel):
    name: str
    command: List[str]
    system: Optional[ProviderDeployResourceLimits] = Field(
        default_factory=ProviderDeployResourceLimits,
    )
    replicas: Optional[int] = 1
    extended_duration: Optional[bool] = True

    settings: Optional[IgnoredType] = IgnoredType()

    # ------------------------------
    # Validators
    # ------------------------------

    @field_validator("command", mode="before")
    def validate_command(cls, value: Any) -> list[str]:
        return validate_command_list(value)

    @field_validator("name", mode="before")
    def validate_name(cls, value: str) -> str:
        return clean_name_string(value)

    @model_validator(mode="before")
    def ensure_system_provider(cls, data, info):
        if not data.get("system"):
            data["system"] = {}

        data["system"]["provider"] = info.context.get("provider")

        return data

    # ------------------------------
    # Properties
    # ------------------------------

    @property
    def is_ext_duration_enabled(self):
        return self.settings.service.environment.is_static and self.extended_duration

    @property
    def joined_command(self):
        return " ".join(self.command)

    # ------------------------------
    # Instance methods
    # ------------------------------

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} {self.name} "
            f"(service: {self.settings.service.name})>"
        )

    __str__ = __repr__


# TODO: discriminator for provider
class ScheduledTask(BaseModel):
    provider: str
    name: str
    command: List[str]
    raw_schedule: str = Field(alias="schedule")
    system: Optional[ProviderDeployResourceLimits] = Field(
        default_factory=ProviderDeployResourceLimits
    )

    # ------------------------------
    # Validators
    # ------------------------------
    @field_validator("command", mode="before")
    def validate_command(cls, value: Any) -> list[str]:
        return validate_command_list(value)

    @field_validator("name", mode="before")
    def validate_name(cls, value: str) -> str:
        return clean_name_string(value)

    @model_validator(mode="before")
    def ensure_system_provider(cls, data, info):
        if not data.get("system"):
            data["system"] = {}

        data["system"]["provider"] = info.context.get("provider")
        data["provider"] = info.context.get("provider")

        return data

    # ------------------------------
    # Properties
    # ------------------------------
    @property
    def schedule(self):
        if self.provider == "gcp":
            return self.raw_schedule
        elif self.provider == "aws":
            cron_exp = self.raw_schedule.split()
            # EventBridge expects an additional year component
            # on cron expressions
            cron_exp.append("*")
            # AWS EventBridge does not allow both day of month
            # and day of week to be specified one of them must be a "?"
            day_of_week = cron_exp[4]
            if day_of_week == "*":
                cron_exp[4] = "?"
            else:
                # day of month
                cron_exp[2] = "?"

            return " ".join(cron_exp)

    @property
    def joined_command(self):
        return " ".join(self.command)


class BackendServiceSettings(BaseServiceSettings):
    type: Literal["backend"]
    seed: List[str] = Field(default_factory=list)
    migrate: List[str] = Field(default_factory=list)
    workers: Optional[List[Worker]] = []
    scheduled_tasks: Optional[List[ScheduledTask]] = []
    cdn: Optional[BackendCDNConfig] = Field(default_factory=BackendCDNConfig)
    integrations: Optional[dict] = None

    # ------------------------------
    # Validators
    # ------------------------------
    @model_validator(mode="after")
    def annotate_children(self):
        for worker in self.workers:
            worker.settings = self

        return self

    @field_validator("seed", mode="before")
    def validate_seed(cls, value: Any) -> list[str]:
        return validate_command_list(value)

    @field_validator("migrate", mode="before")
    def validate_migrate(cls, value: Any) -> list[str]:
        return validate_command_list(value)

    # ------------------------------
    # Properties
    # ------------------------------
    @property
    def is_web(self):
        return True

    @property
    def target_group_name(self):
        if self.service.settings.system.health_check == "/":
            # Only append healthcheck hash
            # if default has been overwritten
            name = f"{self.service.instance_name[-32:].strip('-')}"
        else:
            _healthcheck_hash = hashlib.md5()
            _healthcheck_hash.update(
                (
                    self.service.instance_name
                    + self.service.settings.system.health_check
                ).encode()
            )
            healthcheck_hash = re.sub("\\W|_", "", _healthcheck_hash.hexdigest())
            name = healthcheck_hash[-32:]

        return re.sub("^-+", "", name)



from typing import Optional, List

from .base_model import BaseModel


class CustomHeader(BaseModel):
    name: str
    value: str
    type: Optional[str] = "response"


class CustomHeaders(BaseModel):
    headers: List[CustomHeader] = []
    provider: str



from typing import Literal
from cnc.models.config.deploy_resource_limits import (
    DeployResourceLimits,
)
from cnc.logger import get_logger

log = get_logger(__name__)


class GCPDeployResourceLimits(DeployResourceLimits):
    provider: Literal["gcp"]

    # ------------------------------
    # Properties
    # ------------------------------

    @property
    def cpu(self):
        return self.raw_cpu

    @property
    def memory(self):
        return self.raw_memory

    @property
    def gke_autopilot_cpu(self):
        if self.mem_cpu_ratio > 6.5:
            num_mem_gi = self.normalized_memory
            # mem/cpu comparison is done in Gi/cores
            # => return cpu value in millicores
            return f"{int((num_mem_gi / 6.5) * 1000)}m"

        return self.cpu

    @property
    def gke_autopilot_memory(self):
        if self.mem_cpu_ratio < 1:
            num_cpu_cores = self.normalized_cpu
            # mem/cpu comparison is done in Gi/cores
            # => return memory value in bytes
            return int(num_cpu_cores * (2**30))

        return self.memory

    @property
    def mem_cpu_ratio(self):
        try:
            # (Gi/cores)
            return float(self.normalized_memory / self.normalized_cpu)
        except Exception as e:
            log.warning(
                f"Unable to normalize values for cpu: {self.cpu},"
                f" memory: {self.memory} | error: {e}"
            )

        # Return 1 here if unable to calc ratio so we just use
        # raw values for mem/cpu in that case
        return 1

    @property
    def normalized_cpu(self):
        return self.normalize_cpu_value(self.cpu)

    @property
    def normalized_memory(self):
        return self.normalize_mem_value(self.memory)

    # ------------------------------
    # Instance methods
    # ------------------------------

    def normalize_cpu_value(self, value):
        """
        returns cpu value in cores
        """
        if str(value).endswith("m"):
            return int(value[:-1]) * 1000

        return float(value)

    def normalize_mem_value(self, value):
        """
        returns memory value in Gi
        """
        memory_units = {
            "Ei": 2**60,  # Exbibytes
            "Pi": 2**50,  # Pebibytes
            "Ti": 2**40,  # Tebibytes
            "Gi": 2**30,  # Gibibytes
            "Mi": 2**20,  # Mebibytes
            "Ki": 2**10,  # Kibibytes
            "E": 10**18,  # Exabytes
            "P": 10**15,  # Petabytes
            "T": 10**12,  # Terabytes
            "G": 10**9,  # Gigabytes
            "M": 10**6,  # Megabytes
            "k": 10**3,  # Kilobytes
        }

        for unit, multiplier in memory_units.items():
            if str(value).endswith(unit):
                return float(value[: -len(unit)]) * multiplier / (2**30)



from typing import Optional
from pydantic import model_validator, ValidationInfo, Field
from jinja2 import Template

from .base_model import BaseModel, IgnoredType

from cnc.constants import EnvironmentVariableTypes, EnvironmentVariableDestinations

from cnc.logger import get_logger

log = get_logger(__name__)


class EnvironmentVariable(
    BaseModel, EnvironmentVariableTypes, EnvironmentVariableDestinations
):
    name: Optional[str] = None

    service: Optional[str] = None
    raw_value: Optional[str] = Field(default=None, alias="value")
    raw_secret_id: Optional[str] = Field(default=None, alias="secret_id")
    raw_output_name: Optional[str] = Field(default=None, alias="output_name")
    alias: Optional[str] = None

    collection: Optional[IgnoredType] = Field(default_factory=IgnoredType)
    environment: Optional[IgnoredType] = Field(default_factory=IgnoredType)

    @model_validator(mode="before")
    def validate_existing_info(cls, data: dict, info: ValidationInfo):
        if not (
            data.get("value")
            or data.get("secret_id")
            or data.get("output_name")
            or data.get("alias")
        ):
            raise ValueError(
                "value or secret_id or output_name or alias must be specified"
            )
        return data

    @property
    def secret_id(self):
        if self.variable_type == self.VARIABLE_TYPE_ALIAS:
            for other_variable in self.environment.environment_variables:
                if other_variable.name == self.alias:
                    return other_variable.secret_id
        if self.variable_type == self.VARIABLE_TYPE_SECRET:
            return self.raw_secret_id

    @property
    def output_name(self):
        if self.variable_type == self.VARIABLE_TYPE_ALIAS:
            for other_variable in self.environment.environment_variables:
                if other_variable.name == self.alias:
                    return other_variable.output_name
        if self.variable_type == self.VARIABLE_TYPE_OUTPUT:
            template = Template(self.raw_output_name)
            context = {
                "environment": {"name": self.environment.name},
                "collection": {
                    "name": self.collection.name or self.environment.collection.name
                },
            }
            return template.render(context)

    @property
    def value(self):
        try:
            if self.variable_type == self.VARIABLE_TYPE_SECRET:
                return self.environment.collection.get_secret_value(self.secret_id)
            if self.variable_type == self.VARIABLE_TYPE_STANDARD:
                return self.raw_value
            if self.variable_type == self.VARIABLE_TYPE_OUTPUT:
                return self.environment.collection.get_terraform_output(
                    self.output_name
                )
            if self.variable_type == self.VARIABLE_TYPE_ALIAS:
                for other_variable in self.environment.environment_variables:
                    if other_variable.name == self.alias:
                        return other_variable.value
        except Exception as e:
            log.warning(f"Cannot get value for variable {self.name}: {e}")
            return ""

    @property
    def variable_type(self):
        if self.alias:
            return self.VARIABLE_TYPE_ALIAS
        elif self.raw_secret_id:
            return self.VARIABLE_TYPE_SECRET
        elif self.raw_output_name:
            return self.VARIABLE_TYPE_OUTPUT
        elif self.raw_value:
            return self.VARIABLE_TYPE_STANDARD
        else:
            log.warning(f"Unknown secret type: {self}")



import typer
from IPython import embed
from cnc.models import (
    # Environment,
    # EnvironmentVariable,
    # AppConfig,
    # EnvironmentCollection,
    Application,
)
from .telemetry import send_event

from cnc.logger import get_logger

log = get_logger(__name__)
app = typer.Typer()


def get_app():
    app = Application.from_environments_yml("environments.yml")
    return app


@app.callback(invoke_without_command=True)
@app.command()
def start(ctx: typer.Context):
    send_event("shell.start")
    local_vars = {
        "get_app": get_app,
        "parse_yml": get_app,
    }
    embed(user_ns=local_vars, colors="neutral")
    raise typer.Exit()



from typing import List
import typer

from .telemetry import send_event

from cnc.logger import get_logger

log = get_logger(__name__)


app = typer.Typer()


@app.callback()
def add_common(
    ctx: typer.Context,
    collection_name: str = typer.Option(
        default="preview", envvar="CNC_COLLECTION_NAME", help="Collection name"
    ),
):
    """Common Entry Point for Toolbox"""
    print(ctx.obj)
    collection = ctx.obj.application.collection_by_name(collection_name)
    if not collection:
        log.error(
            f"No collection found for: {collection_name} | (configured "
            f"collections are: {[c.name for c in ctx.obj.application.collections]})"
        )
        raise typer.Exit(code=1)

    ctx.obj.collection = collection


@app.command()
def start(
    ctx: typer.Context,
    environment_name: str,
    service_name: str = "",
    tag: str = "latest",
):
    send_event("toolbox.start")
    collection = ctx.obj.collection

    environment = collection.environment_by_name(environment_name)
    if not environment:
        log.error(f"No environment found for: {environment_name}")
        raise typer.Exit(code=1)

    if service_name:
        service = environment.service_by_name(service_name)
        if not service:
            log.error(f"No service found for: {service_name}")
            raise typer.Exit(code=1)
    else:
        # If no service specified, default to first backend service
        if not environment.backend_services:
            log.error(f"No backend service found for environment: {environment_name}")
            raise typer.Exit(code=1)

        service = environment.backend_services[0]

    toolbox = service.toolbox_manager(environment_tag=tag)
    toolbox.start()

    log.debug(f"Toolbox stopped for {toolbox.config_files_path}")
    raise typer.Exit()


@app.command()
def run(
    ctx: typer.Context,
    environment_name: str,
    command: List[str],
    service_name: str = "",
    tag: str = "latest",
):
    send_event("toolbox.run")
    collection = ctx.obj.collection

    environment = collection.environment_by_name(environment_name)
    if not environment:
        log.error(f"No environment found for: {environment_name}")
        raise typer.Exit(code=1)

    if service_name:
        service = environment.service_by_name(service_name)
        if not service:
            log.error(f"No service found for: {service_name}")
            raise typer.Exit(code=1)
    else:
        # If no service specified, default to first backend service
        if not environment.backend_services:
            log.error(f"No backend service found for environment: {environment_name}")
            raise typer.Exit(code=1)

        service = environment.backend_services[0]

    toolbox = service.toolbox_manager(environment_tag=tag)
    cmd_exit_code = toolbox.run(command)

    log.debug(f"Toolbox stopped for {toolbox.config_files_path}")
    raise typer.Exit(code=cmd_exit_code)



from .base_test_class import CNCBaseTestCase
from cnc.models import Application

from cnc.logger import get_logger

log = get_logger(__name__)


class EnvironmentBaseTestCase(CNCBaseTestCase):
    fixture_name = "backend-1-service-1-db"

    def setUp(self):
        super().setUp()
        app = Application.from_environments_yml("environments.yml")
        self.collection = app.collections[0]
        self.environment = self.collection.environments[0]
        self.assertEqual(self.environment.name, "main")


class EnvironmentUseExistingDBTestCase(EnvironmentBaseTestCase):
    fixture_name = "backend-1-service-1-db-use-existing"

    def test_environment_use_existing_db(self):
        self.assertEqual(len(self.collection.environments), 1)
        self.assertEqual(self.environment.existing_resources[0].name, "db1")
        self.assertEqual(self.environment.existing_resources[0].instance_name, "foobar")


class EnvironmentBothUseExistingDBTestCase(EnvironmentBaseTestCase):
    fixture_name = "backend-1-service-1-db-both-use-existing"

    def test_environment_use_existing_db(self):
        self.assertEqual(len(self.collection.environments), 1)
        self.assertEqual(self.environment.existing_resources[0].name, "db1")
        self.assertEqual(self.environment.existing_resources[0].instance_name, "bazbar")


class EnvironmentCollectionTwoServicesandDBsTwoEnvsTest(EnvironmentBaseTestCase):
    fixture_name = "backend-1-service-2-db-2-envs"

    def test_environment_domain(self):
        self.assertEqual(len(self.collection.environments), 2)
        self.assertEqual(
            self.environment.domain,
            "main.my-backend-test-app.testnewsite.coherencesites.com",
        )

    def test_environment_services(self):
        self.assertEqual(len(self.environment.services), 3)
        self.assertEqual(len(self.environment.web_services), 1)



import os
import typer
import click
from jinja2 import meta
from jinja2.visitor import NodeVisitor

from cnc.models.cycle_stage_base import _TemplatedBase

from cnc.logger import get_logger

log = get_logger(__name__)


app = typer.Typer()


class TemplateNodeInspector(NodeVisitor):
    def __init__(self):
        self.found_extends = []
        self.found_includes = []

    def visit_Extends(self, node):
        self.found_extends.append(node.template.as_const())
        super().generic_visit(node)

    def visit_Include(self, node):
        self.found_includes.append(node.template.as_const())
        super().generic_visit(node)


def find_blocks(node, blocks=None):
    """Recursively find block names in the template AST."""
    if blocks is None:
        blocks = set()

    # Check if the node is a Block node and add it to the blocks set
    if isinstance(node, meta.nodes.Block):
        blocks.add(node.name)

    # If the node has a body attribute, iterate over its contents
    if hasattr(node, "body"):
        # For nodes with a body that is a list (e.g., Template, Block)
        if isinstance(node.body, list):
            for subnode in node.body:
                find_blocks(subnode, blocks)
        # For nodes with a single node as a body (not in a list)
        elif isinstance(node.body, meta.nodes.Node):
            find_blocks(node.body, blocks)

    # Additionally, for nodes representing control structures that have
    # an `else` or `elif` block, we also need to check these blocks.
    if hasattr(node, "elif_"):
        for condition, body in node.elif_:
            find_blocks(body, blocks)
    if hasattr(node, "else_") and node.else_:
        find_blocks(node.else_, blocks)

    return blocks


class TemplateInspector(_TemplatedBase):

    def __init__(self, collection, template_type="build"):
        self.application = collection.application
        self.collection = collection
        self.template_config = self.application.template_config
        self.working_dir = os.getcwd()
        self.template_type = template_type

    def get_parsed_template(self, template_name):
        env = self._template_env()
        template_source = env.loader.get_source(env, template_name)[0]
        return env.parse(template_source)

    def find_variables(self, ast):
        """Find undeclared variables in the template AST."""
        return meta.find_undeclared_variables(ast)

    def find_extends_and_includes(self, ast):
        """Find all extends and includes within an AST."""
        extends = [
            node.template.value
            for node in ast.body
            if node.__class__.__name__ == "Extends"
        ]
        includes = [
            node.template.value
            for node in ast.body
            if node.__class__.__name__ == "Include"
        ]
        return extends, includes

    def analyze_template(self, template_name, analyzed_templates=set()):
        if template_name in analyzed_templates:
            return set()

        analyzed_templates.add(template_name)

        ast = self.get_parsed_template(template_name)
        inspector = TemplateNodeInspector()
        inspector.visit(ast)

        variables = set(self.find_variables(ast))

        # Recursively analyze base templates
        for base_template in inspector.found_extends:
            variables.update(self.analyze_template(base_template, analyzed_templates))

        # Recursively analyze included templates
        for included_template in inspector.found_includes:
            variables.update(
                self.analyze_template(included_template, analyzed_templates)
            )

        return variables

    def inspect_template_context(self):
        _context = {}

        for service in self.collection.all_services_for_type():
            context_kwargs = {}

            if self.template_type == "build":
                from cnc.models import BuildStageManager

                manager = BuildStageManager(self.collection.environments[0])
                context_kwargs = {"service": service}
            elif self.template_type == "deploy":
                from cnc.models import DeployStageManager

                manager = DeployStageManager(self.collection)
                context_kwargs = {"service": service}
            elif self.template_type == "provision":
                from cnc.models.provisioner import ProvisionStageManager

                manager = ProvisionStageManager(self.collection)

            _context[service.name] = manager.template_context(**context_kwargs)
        return _context

    def inspect(self, template_name):
        self.copy_template_dir()
        self.debug_template_output_directory()

        ast = self.get_parsed_template(template_name)
        blocks = find_blocks(ast)
        variables = self.analyze_template(template_name)
        context = self.inspect_template_context()

        print(f"Blocks: {blocks}")
        print(f"Accessed Variables: {variables}")
        print(f"Context: {context}")

        self.cleanup()


@app.command("list")
def list_for_file(
    ctx: typer.Context,
    template_path: str,
    collection_name: str = "",
    template_type: str = typer.Option(
        default="provision", click_type=click.Choice(["build", "provision", "deploy"])
    ),
):
    collection = ctx.obj.application.collection_by_name(collection_name)
    if not collection:
        log.error(f"No collection found for: {collection_name}")
        raise typer.Exit(code=1)

    inspector = TemplateInspector(collection)
    inspector.template_type = template_type
    inspector.inspect(template_path)

    raise typer.Exit()



import unittest
import os
import shutil
import uuid


class CNCBaseTestCase(unittest.TestCase):
    fixture_name = ""
    env_data_filepath = "environments.yml"

    def setUp(self):
        if not self.fixture_name:
            raise ValueError("Must provide fixture_name")

        self.working_dir = f"/tmp/{str(uuid.uuid4())}"

        shared_template_dir = f"{os.path.dirname(__file__)}/fixtures/shared"
        fixtures_path = f"{os.path.dirname(__file__)}/fixtures/{self.fixture_name}"

        shutil.copytree(
            shared_template_dir,
            self.working_dir,
            dirs_exist_ok=True,
        )

        shutil.copytree(
            fixtures_path,
            self.working_dir,
            dirs_exist_ok=True,
        )

        # print(f"\n{self.working_dir}")
        # subprocess.run(["ls", "-al", self.working_dir])
        os.chdir(self.working_dir)
        # subprocess.run(["cat", "cnc.yml"])
        # print("\n")

    def tearDown(self):
        shutil.rmtree(
            self.working_dir,
            ignore_errors=True,
        )



import shutil
import sys

from .logger import get_logger

log = get_logger(__name__)


def check_deps():
    # add provider-specific ones
    # check bash shell available
    # print info of what version of each is installed
    # print tested versions?

    for tool_name in ["terraform"]:
        check_if_installed(tool_name)


def check_if_installed(tool_name):
    if shutil.which(tool_name) is None:
        log.warning(f"Error: The tool '{tool_name}' is not installed.")
        sys.exit(1)



import subprocess

from .cycle_stage_base import EnvironmentTemplatedBase
from .stage import Stage

from cnc.logger import get_logger

log = get_logger(__name__)


class BuildStageManager(EnvironmentTemplatedBase):
    template_type = "build"

    def template_context(self, service):
        return {
            "builder": self,
            "service": service,
            "environment": self.environment,
            "stage": Stage.model_validate({"name": "build"}),
            "load_database_snapshot": False,
        }

    def debug(self):
        self.debug_template_output_directory()

        for script in self.scripts_to_run:
            self.debug_template_output(script)

        for service in self.environment.services:
            self.debug_template_output(f"build-{service.name}-functions.sh")

        log.debug(f"All done with debug for {self}")
        return

    def render_build(self, service_names=None):
        for service in self.environment.services:
            if service_names:
                if service.name not in service_names:
                    continue

            if service.settings.is_resource:
                continue

            context = self.template_context(service)
            context["render_template"] = self.write_template_with_context(service)

            self.write_template(
                "build_functions.sh.j2",
                output_name=f"build-{service.name}-functions.sh",
                context=context,
            )

            service_build_script_name = f"build-{service.name}.sh"
            self.write_template(
                "main.sh.j2", output_name=service_build_script_name, context=context
            )
            self.scripts_to_run.append(service_build_script_name)

    def perform(self, should_cleanup=True, should_regenerate_config=True, debug=False, service_names=None):
        log.debug(f"Performing build for {self} @ {self.config_files_path}")
        if should_cleanup:
            self.cleanup()

        self.setup()

        if should_regenerate_config:
            self.render_build(service_names=service_names)
            log.debug(f"Done rendering build for {self} (svcs: {service_names}), going to execute...")

        if debug:
            self.debug()
            return

        for script in self.scripts_to_run:
            # self.debug_template_output(script)
            # self.debug_template_output(f"build-app-functions.sh")

            try:
                # log.debug(f"Going to run {script}...")
                subprocess.run(
                    ["bash", "-c", f"source {self.rendered_files_path}/{script}"],
                    executable="/bin/bash",
                )
                # log.debug(f"Output from {script}: \n{ret.stdout}\n{ret.stderr}")
            except subprocess.CalledProcessError as e:
                # Handle any errors that occurred during script execution
                print(f"Error running the script: {e}")
                print("Error output:")
                print(e.stderr)

        log.info(f"All done with perform for {self}")
